from PyQt5.QtCore import QVariant
from qgis.core import QgsFeature,QgsField,QgsGeometry,QgsLayerTreeGroup,QgsMapLayer,QgsProject,QgsVectorLayer
from qgis.gui import QgisInterface,QgsMapCanvas
from ..com.orbitgis.tool.message.Message import Message
from ..orbitgt.mapping3d.sdk.viewer.AList import AList
from ..orbitgt.mapping3d.sdk.viewer.MapObject import MapObject
from .MapObjects import MapObjects
from .MapObjectRenderer import MapObjectRenderer

# Class MapObjectLayer manages the display of a list of MapObjects as a layer in QGIS
#
class MapObjectLayer:

    # Create a new layer
    #
    def __init__(self,name: str,iface: QgisInterface):
        self.name = name
        self.iface = iface
        self.mapObjects = AList[MapObject](MapObject.TYPE)
        self.mapRefresh = False
        self.layer = None # QgsVectorLayer
        self.layerId = None
        self.layerRenderer = None # MapObjectRenderer
        QgsProject.instance().layerRemoved.connect(self.layerRemoved)

    # Called by QGIS when a map layer has manually been removed by the user
    #
    def layerRemoved(self,layerId: str):
        if (self.layerId==None): return
        if (layerId==self.layerId):
            Message.print(__name__,"MapObject layer '"+self.name+"' has been removed")
            self.layer = None
            self.layerId = None
    
    # Is a QGIS layer available?
    #
    def hasLayer(self) -> bool:
        return (self.layer!=None);
    
    # Change the visibility of the layer
    #
    def setVisible(self,visible: bool) -> None:
        if (self.layer!=None):
            treeRoot = QgsProject.instance().layerTreeRoot()
            if isinstance(treeRoot,QgsLayerTreeGroup):
                treeLayer = treeRoot.findLayer(self.layer.id())
                if (treeLayer!=None): treeLayer.setItemVisibilityChecked(visible)
        
    # Update the layer to display a new list of map objects (should be called from the main UI thread)
    #
    def update(self,mapObjects: AList[MapObject],mapRefresh: bool) -> None:
        try:
            if (mapObjects==None): return
            self.mapObjects = mapObjects
            self.mapRefresh = mapRefresh
            self._createLayer()
        except Exception as e:
            Message.printError(__name__,e,"Error while updating mapobjects layer")
        
    # Unload the layer (should be called from the main UI thread)
    #
    def unload(self) -> None:
        if (self.layer!=None):
            Message.print(__name__,"Removing layer '"+self.name+"'")
            QgsProject.instance().removeMapLayer(self.layer)
            self.layer = None
            self.layerId = None
    
    # Add all MapObjects to the feature layer
    #
    def _addFeatures(self,mapCanvas: QgsMapCanvas,removeAll: bool) -> None:
        features = []
        for i in range(0,self.mapObjects.size()):
            mapObject = self.mapObjects.get(i)
            geometry = MapObjects.getBoundsAsPolyline(mapObject)
            if (geometry==None):
                Message.printWarning(__name__,"Failed to get a geometry for mapobject")
            if (geometry!=None):
                feature = QgsFeature()
                feature.setGeometry(geometry)
                feature.setFields(self.layer.fields(),True)
                featureId = mapObject.objectId
                if (featureId==None): featureId = str(i)
                feature.setAttribute("id",i)
                feature.setAttribute("MapObjectId",featureId)
                features.append(feature)
        if (removeAll): self.layer.selectAll()
        self.layer.startEditing()
        if (removeAll): self.layer.deleteSelectedFeatures()
        for feature in features: self.layer.addFeature(feature)
        committed = self.layer.commitChanges()
        self.layer.updateExtents()
        Message.print(__name__,"Added "+str(len(features))+" features ("+str(committed)+")")
    
    # Add (or update) the map objects as a (new) layer
    #
    def _createLayer(self) -> None:
        # Do we have map objects?
        if (self.mapObjects==None): return
        if (self.mapObjects.size()==0): return
        # Update the existing layer?
        if (self.layer!=None):
            # Update the renderer
            self.layerRenderer.setMapObjectList(self.mapObjects)
            # Update the features
            self._addFeatures(self.iface.mapCanvas(),True)
            # Repaint?
            if (self.mapRefresh): self.iface.mapCanvas().refresh()
            return
        # Create a new layer?
        if (self.layer==None):
            Message.print(__name__,"Adding layer '"+self.name+"'")
            layerList = QgsProject.instance().mapLayersByName(self.name)
            if not layerList:
                # Create a new vector layer
                self.layer = QgsVectorLayer("linestring?crs=epsg:"+MapObjects.CRS+"&field=id:integer",self.name,"memory")
                self.layerId = self.layer.id()
                self.layer.dataProvider().addAttributes([QgsField("MapObjectId",QVariant.String)])
                self.layer.updateFields()
                if not self.layer.isValid(): raise Exception("Failed to create layer")
                # Disable the "contents will be permanently lost. Are you sure you want to proceed?" warning on closing the project/QGIS
                self.layer.setCustomProperty("skipMemoryLayersCheck",1)
                # Set the custom renderer
                self.layerRenderer = MapObjectRenderer()
                self.layerRenderer.setMapObjectList(self.mapObjects)
                self.layer.setRenderer(self.layerRenderer)
                # Add the features
                self._addFeatures(self.iface.mapCanvas(),False)
                # Add the layer to QGIS
                QgsProject.instance().addMapLayer(self.layer)
                Message.print(__name__,"Added layer '"+self.name+"'")
