from PyQt5.QtGui import QColor
from qgis.core import QgsCoordinateReferenceSystem,QgsCoordinateTransform,QgsFillSymbol,QgsFontMarkerSymbolLayer,QgsGeometry,QgsLineSymbol,QgsMapToPixel,QgsMarkerSymbol,QgsPoint,QgsPointXY,QgsProject,QgsSimpleFillSymbolLayer,QgsSimpleLineSymbolLayer,QgsSimpleMarkerSymbolLayer,QgsSymbol
from qgis.gui import QgsMapCanvas
from ..com.orbitgis.tool.message.Message import Message
from ..orbitgt.mapping3d.sdk.viewer.AList import AList
from ..orbitgt.mapping3d.sdk.viewer.Bounds import Bounds
from ..orbitgt.mapping3d.sdk.viewer.Color import Color
from ..orbitgt.mapping3d.sdk.viewer.Constants import Constants
from ..orbitgt.mapping3d.sdk.viewer.MapObject import MapObject
from .MapObjectSymbolLayer import MapObjectSymbolLayer

class MapObjects():
    # Define the CRS that will be used to communicate with the 3D Mapping SDK (WGS84 lon-lat geographic EPSG:4326)
    CRS: str = Constants.CRS_WGS84
    
    TYPE_AREA: str = "area"
    TYPE_LINE: str = "line"
    TYPE_POINT: str = "point"
    TYPE_SYMBOL: str = "symbol"
    TYPE_OTHER: str = "other"
    
    def getType(mapObject: MapObject) -> str:
        if (mapObject.spatialArea!=None): return MapObjects.TYPE_AREA
        if (mapObject.spatialLine!=None): return MapObjects.TYPE_LINE
        if (mapObject.spatialPoint!=None) and (mapObject.symbolPath==None): return MapObjects.TYPE_POINT
        if (mapObject.spatialPoint!=None) and (mapObject.symbolPath!=None): return MapObjects.TYPE_SYMBOL
        return MapObjects.TYPE_OTHER

    def createColor(color: Color) -> QColor:
        return QColor(color.red,color.green,color.blue,color.alpha)

    def updateSymbol(mapObject: MapObject,symbol: QgsSymbol) -> QgsSymbol:
        if (mapObject.spatialArea!=None) or (mapObject.symbolPath!=None):
            #symbol = QgsFillSymbol.createSimple({})
            symbol.deleteSymbolLayer(0) # Remove default symbol layer
            mapObject.color = Color(0,0,0,255)
            strokeWidth = (mapObject.lineWidth/96.0*25.4)
            symbolLayer = QgsSimpleFillSymbolLayer(color=MapObjects.createColor(mapObject.fillColor),strokeColor=MapObjects.createColor(mapObject.color),strokeWidth=strokeWidth)
            symbol.appendSymbolLayer(symbolLayer)
            return symbol
        if (mapObject.spatialLine!=None):
            #symbol = QgsLineSymbol.createSimple({})
            symbol.deleteSymbolLayer(0) # Remove default symbol layer
            strokeWidth = (mapObject.lineWidth/96.0*25.4)
            symbolLayer = QgsSimpleLineSymbolLayer(color=MapObjects.createColor(mapObject.color),width=strokeWidth)
            symbol.appendSymbolLayer(symbolLayer)
            return symbol
        if (mapObject.spatialPoint!=None):
            #symbol = QgsMarkerSymbol.createSimple({})
            symbol.deleteSymbolLayer(0) # Remove default symbol layer
            symbolLayer = QgsSimpleMarkerSymbolLayer(color=MapObjects.createColor(mapObject.color),size=3.0) # size in MM
            symbol.appendSymbolLayer(symbolLayer)
            return symbol
        return None

    def getPointGeometry(mapObject: MapObject) -> QgsGeometry:
        if (mapObject.spatialPoint==None): return None
        wkt = mapObject.spatialPoint.asWKT()
        Message.print(__name__,"POINT WKT: "+wkt)
        return QgsGeometry.fromWkt(wkt)

    def getLineGeometry(mapObject: MapObject) -> QgsGeometry:
        if (mapObject.spatialLine==None): return None
        wkt = mapObject.spatialLine.asPolylineWKT()
        Message.print(__name__,"LINE WKT: "+wkt)
        return QgsGeometry.fromWkt(wkt)

    def getAreaGeometry(mapObject: MapObject) -> QgsGeometry:
        if (mapObject.spatialArea==None): return None
        wkt = mapObject.spatialArea.asPolygonWKT()
        Message.print(__name__,"AREA WKT: "+wkt)
        return QgsGeometry.fromWkt(wkt)

    def getAnchorPoint(mapObject: MapObject) -> QgsGeometry:
        if (mapObject.spatialPoint!=None):
            point = mapObject.spatialPoint
            return QgsGeometry.fromPointXY(QgsPointXY(point.x,point.y))
        if (mapObject.spatialLine!=None):
            part0 = mapObject.spatialLine.parts.get(0)
            point0 = part0.points.get(0)
            return QgsGeometry.fromPointXY(QgsPointXY(point0.x,point0.y))
        if (mapObject.spatialArea!=None):
            part0 = mapObject.spatialArea.parts.get(0)
            point0 = part0.points.get(0)
            return QgsGeometry.fromPointXY(QgsPointXY(point0.x,point0.y))
        return None

    def getBoundsAsPolyline(mapObject: MapObject) -> QgsGeometry:
        bounds = None
        if (mapObject.spatialPoint!=None):
            bounds = Bounds(mapObject.spatialPoint.crs)
            bounds.addCoordinate(mapObject.spatialPoint)
        if (mapObject.spatialLine!=None):
            bounds = mapObject.spatialLine.getBounds()
        if (mapObject.spatialArea!=None):
            bounds = mapObject.spatialArea.getBounds()
        if (bounds==None): return None
        if (bounds.valid==False): return None
        points = []
        points.append(QgsPointXY(bounds.minX,bounds.minY))
        points.append(QgsPointXY(bounds.maxX,bounds.minY))
        points.append(QgsPointXY(bounds.maxX,bounds.maxY))
        points.append(QgsPointXY(bounds.minX,bounds.maxY))
        points.append(QgsPointXY(bounds.minX,bounds.minY))
        return QgsGeometry.fromPolylineXY(points)

    def getEpsgCode(crs: QgsCoordinateReferenceSystem) -> str:
        if (crs==None): return None
        code = crs.authid()
        if (code==None): return None
        if (code.startswith("EPSG:")): return code[5:]
        return None
